% MATLAB code in support of menke_gls_paper
% Example 1:  GLS implementation for interpolation
% with minimum curvature splines, Bill Menke, Feb 2014
clear all;

% globals necessary to communicate inverse problem
% to function called by biconjugate gradient solver
global G H vard varh

tic

% set up auxiallry variable, position x
Nx=101;
xmin = 0;
xmax = 10;
xL = (xmax-xmin);
Dx = xL/(Nx-1);
x = xmin + Dx*[0:Nx-1]';

% number of model parameters = number of data = Nx
N=Nx;
M=N;

% We want d=Gm to look as much like intergal operation
% d(x0) = (integral) delta(x-x0) m(x) dx as possible
% so G=Dx*I/Dx=I  (first Dx from Reimann Sum, second from delta function
% thus
G=speye(N,N);

% model and data
nw=8;
mtrue=0.75* ones(M,1)+0.25*sin(nw*pi*x/xL);
dtrue = G*mtrue;
sigmad= 0.05;
dobs = dtrue+random('Normal',0,sigmad,N,1);
vard = (sigmad^2)*ones(N,1);

% plot data
figure(1);
clf;
set(gca,'LineWidth',2);
hold on;
axis( [xmin, xmax, 0, 1.5] );
plot( x, dobs, 'ko', 'LineWidth', 2);
xlabel('x');
ylabel('m and d');
title('data: observed (black), fit (red), 95% confidence limits (blue)');

% we want Hm=h to act like a second derivative operator
% h = d2m/dx2 = (integral) delta''(x-x0) m(x) dx
% so it must have rows like
% (Dx^-2)*[... 0, 0, 1, -2, 1, 0, 0, ...]
H = (Dx^-2)*spdiags( [ones(N,1),-2*ones(N,1),ones(N,1)], [0,1,2], N-2, M );
sigmah=10;
sigmah=1;
varh = (sigmah^2)*ones(N-2,1);

% generalized least squares with equations Gm=d
% with covariance Cd = diag(vard)
% and prior information Hm=h=0
% with covariance Ch = diag(varh)
% generalized least squares solution
% mest = inv(G'*inv(Cd)*G + H'*inv(Ch)*H)) * Gmg * inv(Cd) * dobs
% solve by biconjugate gradients
mest = bicg( @glsfcn, G'*(dobs./vard), 1e-5, 4*N );
plot( x, mest, 'r-', 'LineWidth', 3);

% check against simple method ... check OK
% Cdi = diag( ones(N,1)./vard );
% Chi = diag( ones(N-2,1)./varh );
% mest2 = (G'*Cdi*G + H'*Chi*H)\(G'*Cdi*dobs);
% plot( x, mest2, 'b:', 'LineWidth', 3);

% error bar calculation
varlist=[1:floor(N/6):N];
for k = varlist
    % compute column vk of inverse of A=(G'*Cdi*G + H'*Chi*H)
    % note inv(A) symmetric, so column is also row
    s = zeros(M,1);
    s(k)=1;
    vk = bicg( @glsfcn, s, 1e-5, 4*N );

    % check inv(A) via simpler method ... Check OK
    % Cdi = diag( ones(N,1)./vard );
    % Chi = diag( ones(N-2,1)./varh );
    % AI=inv(G'*Cdi*G + H'*Chi*H);
    % vk_check = AI(:,k);
    % [vk,vk_check]

    % variance of estimated model parameters is k-th element of vk
    sigmamest = sqrt(vk(k));
    % 95% confidence interval
    mlow = mest(k)-2*sigmamest;
    mhigh = mest(k)+2*sigmamest;
    plot( [x(k), x(k)], [mlow, mhigh], 'b-', 'LineWidth', 3 );

end

% resolution calculation
reslist=[1:floor(N/6):N];
reslist=[1:floor(N/6):N];
Nres=length(reslist);
i=0;
figure(2);
clf;
for k = reslist
    
    % column of the resolution matrix
    % spike model
    s = zeros(M,1);
    s(k)=1;
    % corresponding spiked data
    ds = G*s;
    % solve for column r of resolution matrix
    r = bicg( @glsfcn, G'*(ds./vard), 1e-5, 4*N );
    % but, in this case, R is symmetric so column is row
    
    % plot the column of the resolution matrix
    subplot(Nres+1,1,i+2);
    set(gca,'LineWidth',2);
    hold on;
    axis( [xmin, xmax, -1.1*max(r), 1.1*max(r) ] );
    plot( [x(k), x(k)], [0, 1], 'k-', 'LineWidth', 2 );
    plot( x, r, 'g-', 'LineWidth', 4 );
    
    % check against simple method ... check OK
    % Cdi = diag( ones(N,1)./vard );
    % Chi = diag( ones(N-2,1)./varh );
    % rcheck = (G'*Cdi*G + H'*Chi*H)\(G'*Cdi*ds);
    % plot( x, rcheck, 'b:', 'LineWidth', 3);
    
    % row of the resolution matrix
    % first: compute column of inverse of A=(G'*Cdi*G + H'*Chi*H)
    % note inv(A) symmetric, so column is also row
    vk = bicg( @glsfcn, s, 1e-5, 4*N );
    % second: row of generalized inverse GMG(k,:)=gi
    gi = ((vk')*(G'))./(vard');
    % check GMG) via simpler method ... Check OK
    % Cdi = diag( ones(N,1)./vard );
    % Chi = diag( ones(N-2,1)./varh );
    % GMG=inv(G'*Cdi*G + H'*Chi*H)*G'*Cdi;
    % gi_check = GMG(k,:);
    % [gi',gi_check']
    % third: row of the resolution matrix
    r_row = gi*G;
    % check inv(A) via simpler method
    % also check that its symmetric ... Check OK
    % Cdi = diag( ones(N,1)./vard );
    % Chi = diag( ones(N-2,1)./varh );
    % Rcheck=inv(G'*Cdi*G + H'*Chi*H)*G'*Cdi*G;
    % r_row_check = Rcheck(k,:);
    % r_col_check = Rcheck(:,k);
    % [r_row',r_col_check,r_row_check']
    plot( x, r, 'b-', 'LineWidth', 3 );
    
    % analytic solution based on solving equivalent
    % differential equaltion
    gamma = sigmad/sigmah;
    x0 = x(k);
    a=sqrt(2*gamma);
    A = Dx*(a^3)/(8*(gamma^2));
    r2 = A*exp(-abs(x-x0)/a).*(cos(abs(x-x0)/a)+sin(abs(x-x0)/a));
    plot( x, r2, 'r-', 'LineWidth', 2 );
    i=i+1;
end

toc